from datetime import (
    datetime,
)
from logging import (
    Logger,
)
import traceback
from typing import (
    Any,
    Sequence,
)

from eth_typing import (
    BlockNumber,
)

from ethhelper.datatypes.eth import (
    Address,
    Block,
    FilterParams,
    Hash32,
    Log,
)
from ethhelper.datatypes.geth import (
    GethError,
)

from .eth import (
    GethEthHttp,
)
from .net import (
    GethNetHttp,
)
from .txpool import (
    GethTxpoolHttp,
)


class GethCustomHttp(GethEthHttp, GethNetHttp, GethTxpoolHttp):
    """A customized HTTP interface for Geth nodes that inherits from
    ``GethEthHttp``, ``GethNetHttp``, and ``GethTxpoolHttp``. Provides
    additional functionalities to access Geth nodes via HTTP.

    The ``url`` is used to indicate the path of the HTTP service of the Geth
    node, usually in the form of ``http://host:port/``. The use of third-party
    nodes may be out of the ordinary.

    The ``logger`` is used to assign a logger of the Python logging module to
    this class. Explicitly assigning a logger can be used to control the output
    location of the logger, which is convenient for debugging.
    """
    def __init__(self, url: str, logger: Logger) -> None:
        super().__init__(url, logger)

    async def test_connection(self) -> bool:
        """Test connectivity to the Geth node via both Web3.py and the
        customized HTTP interface.

        Returns:
            A bool indicating whether the connection to the Geth node is
            successful or not.
        """
        try:
            web3 = await self.is_connected()
            customized = await super(GethTxpoolHttp, self).is_connected()
            self.logger.info(
                f"GethHttpWeb3: {web3}, GethTxpoolHttp: {customized}."
            )
            return web3 and customized
        except Exception:
            self.logger.error("GethCustomHttp can't be running.")
            self.logger.debug(f"Detail: {traceback.format_exc()}")
            return False
    
    async def get_logs(self, filter: FilterParams) -> list[Log]:
        """Retrieve a list of logs from the Geth node using the given
        ``FilterParams``.

        Args:
            filter: A FilterParams object used to specify filter parameters for
                the logs.

        Returns:
            A list of Log objects parsed from the logs returned by the Geth
            node.

        Raises:
            ethhelper.types.GethError: Raised when the Geth node returns an
                error.
        """
        async_filter = await self.eth_filter(filter)
        logs = await async_filter.get_all_entries()
        await self.eth_uninstall_filter(async_filter)
        return [Log.parse_obj(log) for log in logs]

    async def get_logs_by_blocks(
        self,
        start_height: BlockNumber,
        end_height: BlockNumber,
        address: Address | list[Address] | None = None,
        topics: Sequence[Hash32 | Sequence[Hash32]] | None = None
    ) -> list[Log]:
        """Retrieve a list of logs within a range of blocks specified by block
        heights.

        Args:
            start_height: The block height to start retrieving logs from.
            end_height: The block height to stop retrieving logs from.
            address: An address or list of addresses to filter the logs by.
            topics: A list of topics or nested lists of topics to filter the
                logs by.

        Returns:
            A list of Log objects parsed from the logs returned by the Geth
            node.

        Raises:
            ethhelper.types.GethError: Raised when the Geth node returns an
                error.
        """
        if end_height - start_height > 200:
            results: list[Log] = []
            for i in range(start_height, end_height + 1, 201):
                if i + 200 > end_height:
                    results += await self.get_logs_by_blocks(
                        BlockNumber(i), end_height, address, topics
                    )
                else:
                    results += await self.get_logs_by_blocks(
                        BlockNumber(i), BlockNumber(i + 200), address, topics
                    )
            return results
        else:
            fliter_params = FilterParams(  # type: ignore
                address = address,
                from_block = start_height,  # type: ignore
                to_block = end_height,  # type: ignore
                topics = topics
            )
            return await self.get_logs(fliter_params)

    async def _binary_search(
        self, start: BlockNumber, end: BlockNumber, target: int
    ) -> BlockNumber:
        """
        Perform binary search to find the block height which is closest to the
        target timestamp.

        Args:
            start: The start block height to search from.
            end: The end block height to search to.
            target: The target timestamp to search for.

        Returns:
            A ``BlockNumber`` instance that represents the block height which
            is closest to the target timestamp.
        """
        middle = (start + end) // 2
        ts = (await self.eth_get_block(BlockNumber(middle))).timestamp
        if end - start <= 1:
            if ts >= target:
                return start
            else:
                return end
        if ts == target:
            return BlockNumber(middle)
        if ts > target:
            return await self._binary_search(
                start, BlockNumber(middle), target
            )
        else:
            return await self._binary_search(BlockNumber(middle), end, target)

    async def get_height_after_ts(self, timestamp: int) -> BlockNumber:
        """
        Get the block height closest to the target timestamp.

        Args:
            timestamp: The target timestamp in seconds since the epoch.

        Returns:
            A ``BlockNumber`` instance that represents the block height closest
            to the target timestamp.
        """
        height_now: int = await self.eth_block_number()
        ts_now = int(datetime.now().timestamp())
        td = ts_now - timestamp
        pos_lower = height_now - (td // 12) - 5
        pos_upper = height_now - (td // 24) + 5
        return await self._binary_search(
            BlockNumber(pos_lower), BlockNumber(pos_upper), timestamp
        )
    
    async def get_blocks_by_numbers(
        self, start: BlockNumber, end: BlockNumber
    ) -> list[Block]:
        """
        Get the blocks with the given block numbers.

        Args:
            start: The start block number.
            end: The end block number.

        Returns:
            A list of ``Block`` instances that represent the blocks with the
            given block numbers.
        """
        if end - start > 200:
            results: list[Block] = []
            for i in range(start, end + 1, 201):
                if i + 200 > end:
                    results += await self.get_blocks_by_numbers(
                        BlockNumber(i), end
                    )
                else:
                    results += await self.get_blocks_by_numbers(
                        BlockNumber(i), BlockNumber(i + 200)
                    )
            return results
        else:
            requests: list[tuple[str, list[Any] | None]] = []
            for i in range(start, end + 1):
                requests.append(("eth_getBlockByNumber", [hex(i), False]))
            success, errors = await self.send_multiple(requests)
            if len(errors) != 0:
                raise GethError(error=[err.error for err in errors])
            return [Block.parse_obj(suc.result) for suc in success]



import typing
from typing import (
    NewType,
    Sequence,
)

from eth_typing import (
    BlockNumber,
)
import orjson
from pydantic import (
    BaseModel,
    Field,
)
from web3.types import (
    ENS,
    FilterParams as Web3FilterParams,
    Nonce,
    TxParams as Web3TxParams,
)

from ethhelper.utils import (
    convert,
    json,
)

from .base import (
    Address,
    BlockIdentifier,
    Gas,
    Hash32,
    HexBytes,
    IntStr,
    Wei,
)


class AccessEntry(BaseModel):
    """An entry representing an address with associated storage keys in an
    access list.
    """
    address: Address
    """The address of the entry."""
    storage_keys: list[Hash32] = Field(alias="storageKeys")
    """The storage keys associated with the entry."""

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        frozen = True
        json_loads = orjson.loads
        json_dumps = json.orjson_dumps


AccessList = NewType("AccessList", Sequence[AccessEntry])
"""A new type representing a sequence of ``AccessEntry`` instances."""


class SyncStatus(BaseModel):
    """The current synchronization status of the Ethereum node."""
    current_block: int = Field(alias="currentBlock")
    """The current block number."""
    highest_block: int = Field(alias="highestBlock")
    """The highest block number seen."""
    starting_block: int = Field(alias="startingBlock")
    """The block number at which the node started syncing."""

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        frozen = True
        json_loads = orjson.loads
        json_dumps = json.orjson_dumps


class FeeHistory(BaseModel):
    """A record of base fees and gas usage ratios."""
    base_fee_per_gas: list[Wei] = Field(alias="baseFeePerGas")
    """A list of base fees per gas for each block in the history."""
    gas_used_ratio: list[float] = Field(alias="gasUsedRatio")
    """A list of gas used ratios for each block in the history."""
    oldest_block: BlockNumber = Field(alias="oldestBlock")
    """The block number of the oldest block in the history."""
    reward: list[list[Wei]] | None
    """A nested list of miner rewards for each block in the history, or
    ``None`` if rewards are not available.
    """

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        frozen = True
        json_loads = orjson.loads
        json_dumps = json.orjson_dumps


class TxParams(BaseModel):
    """Represents a set of parameters required for creating a transaction."""
    chain_id: int | None = Field(None, alias="chainId")
    """The chain ID for the transaction."""
    data: HexBytes | None = None
    """The data payload of the transaction."""
    from_: Address | ENS | None = Field(None, alias="from")
    """The sender's address or ENS name."""
    gas: Gas | None = None
    """The maximum amount of gas to use for the transaction."""
    # legacy pricing
    gas_price: Wei | None = Field(None, alias="gasPrice")
    """The price of gas for the transaction. (Legacy pricing)"""
    # dynamic fee pricing
    max_fee_per_gas: Wei | None = Field(None, alias="maxFeePerGas")
    """The maximum fee per gas for the transaction. (Dynamic fee pricing)"""
    max_priority_fee_per_gas: Wei | None = \
        Field(None, alias="maxPriorityFeePerGas")
    """The maximum priority fee per gas for the transaction. (Dynamic fee
    pricing)
    """
    nonce: Nonce | None = None
    """A unique number used to prevent replay attacks."""
    # addr or ens
    to: Address | ENS
    """The recipient's address or ENS name."""
    type: int | None = None
    """The transaction type, if the chain supports EIP-2718."""
    value: Wei | None = None
    """The amount of Ether to send along with the transaction."""

    def to_web3(self) -> Web3TxParams:
        """Convert the transaction parameters to ``web3.types.TxParams``
        format.

        Returns:
            A dictionary representing the transaction parameters in
            ``web3.types.TxParams`` format.
        """

        td = self.dict(by_alias=True, exclude_none=True)
        if isinstance(td["to"], Address):
            td["to"] = td["to"].to_web3()
        if "data" in td:
            td["data"] = typing.cast(HexBytes, td["data"]).value
        if "from" in td:
            if isinstance(td["from"], Address):
                td["from"] = td["from"].to_web3()
        if "gasPrice" in td:
            td["gasPrice"] = typing.cast(Wei, td["gasPrice"]).to_web3()
        if "maxFeePerGas" in td:
            td["maxFeePerGas"] = typing.cast(Wei, td["maxFeePerGas"]).to_web3()
        if "maxPriorityFeePerGas" in td:
            td["maxPriorityFeePerGas"] = \
                typing.cast(Wei, td["maxPriorityFeePerGas"]).to_web3()
        if "value" in td:
            td["value"] = typing.cast(Wei, td["value"]).to_web3()
        return typing.cast(Web3TxParams, td)

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        json_loads = orjson.loads
        json_dumps = json.orjson_dumps


class Transaction(BaseModel):
    """A class representing an Ethereum transaction."""
    chain_id: int = Field(1, alias="chainId")
    """The chain ID for the transaction."""
    block_hash: Hash32 | None = Field(alias="blockHash")
    """The hash of the block containing the transaction."""
    block_number: BlockNumber | None = Field(alias="blockNumber")
    """The block number containing the transaction."""
    transaction_index: int | None = Field(alias="transactionIndex")
    """The transaction index within the block."""
    hash: Hash32
    """The hash of the transaction."""
    from_: Address = Field(alias="from")
    """The address of the sender of the transaction."""
    to: Address | None
    """The address of the recipient of the transaction (if applicable)."""
    gas: Gas
    """The gas limit for the transaction."""
    gas_price: Wei = Field(alias="gasPrice")
    """The gas price for the transaction."""
    max_fee_per_gas: Wei | None = Field(None, alias="maxFeePerGas")
    """The maximum fee per gas for the transaction (if applicable)."""
    max_priority_fee_per_gas: Wei | None = Field(
        None, alias="maxPriorityFeePerGas"
    )
    """The maximum priority fee per gas for the transaction (if applicable)."""
    input: HexBytes
    """The input data for the transaction."""
    value: Wei
    """The value of the transaction in Wei."""
    nonce: Nonce
    """The nonce of the transaction."""
    type: int
    """The type of the transaction (0 for legacy, 1 for EIP-2930, 2 for
    EIP-1559).
    """
    access_list: AccessList | None = Field(None, alias="accessList")
    """The access list for the transaction (if applicable)."""
    r: HexBytes
    """The r value of the transaction signature."""
    s: HexBytes
    """The s value of the transaction signature."""
    v: int
    """The v value of the transaction signature."""
    # vaildators
    int_val = convert.int_validator(
        "block_number", "gas", "gas_price", "max_fee_per_gas", "v",
        "max_priority_fee_per_gas", "nonce", "value", "type", "chain_id"
    )

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        frozen = True
        json_loads = orjson.loads
        json_dumps = json.orjson_dumps


class Block(BaseModel):
    difficulty: IntStr
    """The difficulty (in hashes) of mining this block. After the Merge
    (15537394), this value will be always 0.
    """
    extra_data: HexBytes = Field(alias="extraData")
    """Additional data provided by the miner. After the Merge (15537394), this
    value will be always empty bytes.
    """
    gas_limit: Gas = Field(alias="gasLimit")
    """The maximum amount of gas that was available to transactions in this
        block.
    """
    gas_used: Gas = Field(alias="gasUsed")
    """The amount of gas that was used to execute transactions in this block.
    """
    base_fee_per_gas: Wei | None = Field(None, alias="baseFeePerGas")
    """The fee per unit of gas burned by the protocol in this block. Before
        EIP-1559 (12965000), this value was always None.
    """
    logs_bloom: HexBytes = Field(alias="logsBloom")
    """A bloom filter that can be used to check if a block may contain log
        entries matching a filter.
    """
    miner: Address
    """The address (20 bytes) of the account that mined this block."""
    nonce: HexBytes
    """The block nonce, a sequence (8 bytes) determined by the miner."""
    number: BlockNumber
    """The number of this block, starting at 0 for the genesis block."""
    parent_hash: Hash32 = Field(alias="parentHash")
    """The hash (32 bytes) of the parent block of this block."""
    hash: Hash32
    """The hash (32 bytes) of this block."""
    mix_hash: Hash32 = Field(alias="mixHash")
    """The hash (32 bytes) that was used as an input to the PoW process. After
        the Merge (15537394), this value will be the beacon chain's RANDAO
        value.
    """
    receipts_root: HexBytes = Field(alias="receiptsRoot")
    """The keccak256 hash (32 bytes) of the trie of transaction receipts in
        this block.
    """
    sha3_uncles: HexBytes = Field(alias="sha3Uncles")
    """The keccak256 hash (32 bytes) of all the ommers (AKA uncles) associated
        with this block.
    """
    state_root: HexBytes = Field(alias="stateRoot")
    """The keccak256 hash (32 bytes) of the state trie after this block was
        processed.
    """
    timestamp: int
    """The unix timestamp at which this block was mined."""
    transactions_root: HexBytes = Field(alias="transactionsRoot")
    """The keccak256 hash (32 bytes) of the root of the trie of transactions in
        this block.
    """
    size: int | None = None
    """The block size in bytes"""
    total_difficulty: IntStr | None = Field(None, alias="totalDifficulty")
    """The sum of all difficulty values up to and including this block. After
        the Merge (15537394), this value is always 58750003716598352816469.
    """
    transactions: list[Hash32] | list[Transaction] | None = None
    """A list of the hash (32 bytes) of transactions associated with this
        block.
    """
    uncles: list[Hash32] | None = None
    """A list of hash of the uncle blocks of this block. After the Merge
    (15537394), this value will be always None.
    """
    # vaildators
    int_val = convert.int_validator(
        "difficulty", "gas_limit", "gas_used", "base_fee_per_gas", "number",
        "timestamp", "size", "total_difficulty"
    )

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        frozen = True
        json_loads = orjson.loads
        json_dumps = json.orjson_dumps


class Log(BaseModel):
    """A published Ethereum event log."""
    block_number: BlockNumber = Field(alias="blockNumber")
    """The number of the block in which the log was generated."""
    block_hash: Hash32 = Field(alias="blockHash")
    """The hash of the block in which the log was generated."""
    log_index: int = Field(alias="logIndex")
    """The index of this log in the block."""
    address: Address
    """The address of the contract that generated this log."""
    topics: list[Hash32]
    """A list of up to 4 32-byte indexed topics for the log."""
    data: HexBytes
    """The unindexed data for this log."""
    transaction_hash: Hash32 = Field(alias="transactionHash")
    """The hash (32 bytes) of the transaction that generated this log."""
    transaction_index: int = Field(alias="transactionIndex")
    """The index of the transaction in the block."""
    removed: bool
    """Whether or not the log was removed from the blockchain."""
    # vaildators
    int_val = convert.int_validator(
        "block_number", "log_index", "transaction_index"
    )

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        frozen = True
        json_loads = orjson.loads
        json_dumps = json.orjson_dumps


class Receipt(BaseModel):
    """A transaction receipt."""
    block_number: BlockNumber = Field(alias="blockNumber")
    """The number of the block this receipt is associated with."""
    block_hash: Hash32 = Field(alias="blockHash")
    """The hash of the block this receipt is associated with."""
    contract_address: Address | None = Field(alias="contractAddress")
    """The address of the contract created by this transaction if it is a
    contract creation transaction, otherwise None.
    """
    cumulative_gas_used: Gas = Field(alias="cumulativeGasUsed")
    """The total amount of gas used in the block up to and including this
    transaction.
    """
    effective_gas_price: Wei = Field(alias="effectiveGasPrice")
    """The amount of ETH paid per unit of gas in the transaction, including the
    amount that is burned by EIP-1559.
    """
    from_: Address = Field(alias="from")
    """The address of the account that sent this transaction."""
    gas_used: Gas = Field(alias="gasUsed")
    """The amount of gas used by this transaction."""
    logs: list[Log]
    """The logs emitted by this transaction."""
    logs_bloom: HexBytes = Field(alias="logsBloom")
    """A bloom filter of the logs emitted by this transaction."""
    status: int
    """The status code of this transaction, where 0 represents success and
    non-zero represents failure."""
    to: Address
    """The address of the account or contract that received this transaction.
    """
    transaction_hash: Hash32 = Field(alias="transactionHash")
    """The hash of the transaction that generated this receipt."""
    transaction_index: int = Field(alias="transactionIndex")
    """The index of this transaction in the block."""
    type: int
    """The type of this transaction, where 0 represents a legacy transaction,
    1 represents an EIP-2930 transaction, and 2 represents an EIP-1559
    transaction.
    """
    # vaildators
    int_val = convert.int_validator(
        "block_number", "cumulative_gas_used", "effective_gas_price",
        "gas_used", "status", "transaction_index", "type"
    )

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        frozen = True
        json_loads = orjson.loads
        json_dumps = json.orjson_dumps


class FilterParams(BaseModel):
    """Parameters used for creating Ethereum filters."""
    address: Address | list[Address] | None = None
    """An Ethereum address or list of addresses to filter on. If None, all
        addresses are filtered.
    """
    block_hash: Hash32 | None = Field(None, alias="blockHash")
    """A block hash (32 bytes) to filter on. If None, all blocks are filtered.
    """
    from_block: BlockIdentifier | None = Field(None, alias="fromBlock")
    """The starting block for filtering. If None, the filter will start from
        the earliest block.
    """
    to_block: BlockIdentifier | None = Field(None, alias="toBlock")
    """The ending block for filtering. If None, the filter will end at the
        latest block.
    """
    topics: Sequence[Hash32 | Sequence[Hash32]] | None = None
    """A list of topics (32 bytes) to filter on. Each topic can be a single
        value or a list of values. If None, all topics are filtered.
    """

    def to_web3(self) -> Web3FilterParams:
        """Converts the ``FilterParams`` object to a
        ``web3.types.FilterParams`` object.
        """
        td = self.dict(by_alias=True, exclude_none=True)
        if "address" in td:
            if isinstance(td["address"], Address):
                td["address"] = td["address"].to_web3()
            else:
                td["address"] = typing.cast(list[Address], td["address"])
                td["address"] = [addr.to_web3() for addr in td["address"]]
        if "blockHash" in td:
            td["blockHash"] = str(typing.cast(Hash32, td["blockHash"]))
            if "fromBlock" in td or "toBlock" in td:
                raise ValueError(
                    "You should only choose one of blockHash or "
                    "fromBlock/toBlock as filter params"
                )
        if "fromBlock" in td:
            td["fromBlock"] = convert.block_id_transfer(td["fromBlock"])
            if "toBlock" not in td:
                td["toBlock"] = "latest"
        if "toBlock" in td:
            td["toBlock"] = convert.block_id_transfer(td["toBlock"])
            if "fromBlock" not in td:
                td["fromBlock"] = "latest"
        if "topics" in td and len(td["topics"]) != 0:
            td["topics"] = typing.cast(
                Sequence[Hash32 | Sequence[Hash32]], td["topics"]
            )
            if isinstance(td["topics"][0], Sequence):
                td["topics"] = typing.cast(
                    Sequence[Sequence[Hash32]], td["topics"]
                )
                td["topics"] = [
                    [str(top2) for top2 in top1] for top1 in td["topics"]
                ]
            else:
                td["topics"] = typing.cast(Sequence[Hash32], td["topics"])
                td["topics"] = [str(top) for top in td["topics"]]
        return typing.cast(Web3FilterParams, td)

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        json_loads = orjson.loads
        json_dumps = json.orjson_dumps



import logging
from logging import (
    Logger,
)

from .base import (
    GethHttpAbstract,
    GethHttpCustomized,
    GethHttpWeb3,
)
from .custom import (
    GethCustomHttp,
)
from .eth import (
    GethEthHttp,
)
from .net import (
    GethNetHttp,
)
from .txpool import (
    GethTxpoolHttp,
)


class GethHttpConnector(GethCustomHttp):
    """``GethHttpConnector`` is an asynchronous wrapper for all HTTP interfaces
    supported by ETHHelper.

    Due to the flat design style, all interfaces use a simple naming structure
    as the member methods of this class.

    The ``url`` is used to indicate the path of the HTTP service of the Geth
    node, usually in the form of ``http://host:port/``. The use of third-party
    nodes may be out of the ordinary.

    The ``logger`` is used to assign a logger of the Python logging module to
    this class. Explicitly assigning a logger can be used to control the output
    location of the logger, which is convenient for debugging. If this
    parameter is not provided or ``None`` is giving, this class will
    automatically call ``logging.getLogger("GethHttpConnector")`` to generate a
    default logger.
    """
    def __init__(self, url: str, logger: Logger | None = None) -> None:
        if logger is None:
            logger = logging.getLogger("GethHttpConnector")
        super().__init__(url, logger)


__all__ = [
    "GethHttpAbstract",
    "GethHttpCustomized",
    "GethHttpWeb3",
    "GethCustomHttp",
    "GethEthHttp",
    "GethNetHttp",
    "GethTxpoolHttp",
    "GethHttpConnector"
]



from logging import (
    Logger,
)
from typing import (
    Literal,
)

from eth_typing import (
    BlockNumber,
)
from web3._utils.filters import (
    AsyncFilter,
)
from web3.eth.async_eth import (
    AsyncEth,
)
from web3.types import (
    ENS,
    BlockParams,
    CallOverride as Web3CallOverride,
    Nonce,
)

from ethhelper.datatypes.base import (
    Address,
    BlockIdentifier,
    Gas,
    Hash32,
    HexBytes,
    Wei,
)
from ethhelper.datatypes.eth import (
    Block,
    FeeHistory,
    FilterParams,
    Receipt,
    SyncStatus,
    Transaction,
    TxParams,
)
from ethhelper.datatypes.geth import (
    CallOverride,
)
from ethhelper.utils import (
    convert,
)

from .base import (
    GethHttpWeb3,
)


class GethEthHttp(GethHttpWeb3):
    """Class for interacting with an Ethereum node using HTTP with the Geth
    JSON-RPC API.

    This class extends the GethHttpWeb3 class and provides additional methods
    specifically for interacting with the Ethereum network using the Geth
    JSON-RPC API.

    The ``url`` is used to indicate the path of the HTTP service of the Geth
    node, usually in the form of ``http://host:port/``. The use of third-party
    nodes may be out of the ordinary.

    The ``logger`` is used to assign a logger of the Python logging module to
    this class. Explicitly assigning a logger can be used to control the output
    location of the logger, which is convenient for debugging.
    """
    def __init__(self, url: str, logger: Logger) -> None:
        super().__init__(url, logger)
        self.eth: AsyncEth = self.w3.eth
        """Asynchronous Eth interface for Web3. Used to simplify the access
        path.
        """

    async def eth_accounts(self) -> list[Address]:
        """Get a list of all accounts in the currently connected Ethereum node.

        Returns:
            A list containing the addresses of all accounts in the current
            node.
        """
        return [Address(addr) for addr in await self.eth.accounts]

    async def eth_hashrate(self) -> int:
        """Get the hash rate of the current Ethereum node.

        Returns:
            The hash rate of the Ethereum node.
        """
        return await self.eth.hashrate

    async def eth_block_number(self) -> BlockNumber:
        """Get the latest block number on the current Ethereum node.

        Returns:
            The latest block number on the current Ethereum node.
        """
        return await self.eth.block_number

    async def eth_chain_id(self) -> int:
        """Get the chain ID of the current Ethereum node.
    
        Returns:
            The chain ID of the current Ethereum node.
        """
        return await self.eth.chain_id

    async def eth_gas_price(self) -> Wei:
        """Get the latest average gas price on the current Ethereum node.
    
        Returns:
            The latest average gas price on the current Ethereum node.
        """
        return Wei(await self.eth.gas_price)

    async def eth_max_priority_fee_per_gas(self) -> Wei:
        """Get the latest maximum priority fee per gas unit on the current
        Ethereum node.
    
        Returns:
            The latest maximum priority fee per gas unit on the current
            Ethereum node.
        """
        return Wei(await self.eth.max_priority_fee)

    async def eth_mining(self) -> bool:
        """Check whether the current Ethereum node is mining.

        Returns:
            ``True`` if the current Ethereum node is mining, ``False``
            otherwise.
        """
        return await self.eth.mining

    async def eth_syncing(self) -> SyncStatus | bool:
        """Check whether the current Ethereum node is syncing the blockchain.

        Returns:
            If the current Ethereum node is syncing the blockchain, returns a
            ``SyncStatus`` object; otherwise, returns ``False`` if it has
            completed syncing.
        """
        result = await self.eth.syncing
        if isinstance(result, bool):
            return result
        return SyncStatus.parse_obj(result)

    async def eth_fee_history(
        self,
        block_count: int,
        newest_block: BlockParams | BlockNumber,
        reward_percentiles: list[float] | None = None,
    ) -> FeeHistory:
        """Retrieve historical fee data for a specified number of the latest
        blocks.
    
        Args:
            block_count: The number of blocks to retrieve.
            newest_block: The parameter or block number of the latest block.
            reward_percentiles: The list of fee percentile values to retrieve.
                Defaults to ``None``.
        
        Returns:
            A ``FeeHistory`` object containing the historical fee data.
        """
        return FeeHistory.parse_obj(
            await self.eth.fee_history(
                block_count, newest_block, reward_percentiles
            )
        )

    async def eth_call(
        self,
        transaction: TxParams,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride = {},
        ccip_read_enabled: bool = False,
    ) -> HexBytes:
        """Execute a read-only transaction (which won't change the blockchain
        state) and return the result.

        Args:
            transaction: The transaction parameters.
            block_identifier: The block identifier. Defaults to "latest".
            state_override: The state override object. Defaults to {}.
            ccip_read_enabled: Whether to enable CCIP read. Defaults to
                ``False``.

        Returns:
            The result of the executed transaction as a ``HexBytes`` object.
        """
        block_id = convert.block_id_transfer(block_identifier)
        state_over: Web3CallOverride = {}
        for addr in state_override:
            state_over[addr] = state_override[addr].to_web3()
        return HexBytes(
            (
                await self.eth.call(
                    transaction.to_web3(),
                    block_id,
                    state_over,
                    ccip_read_enabled
                )
            ).hex()
        )

    async def eth_estimate_gas(
        self,
        transaction: TxParams,
        block_identifier: BlockIdentifier = "latest"
    ) -> Gas:
        """Estimate the gas cost of a transaction.

        Args:
            transaction: The transaction parameters.
            block_identifier: The block identifier. Defaults to "latest".

        Returns:
            The estimated gas cost of the transaction as a ``Gas`` object.
        """
        block_id = convert.block_id_transfer(block_identifier)
        return Gas(
            await self.eth.estimate_gas(transaction.to_web3(), block_id)
        )

    async def eth_get_transaction(
        self, transaction_hash: Hash32
    ) -> Transaction:
        """Get information about a specific transaction by its hash.

        Args:
            transaction_hash: The hash of the transaction to retrieve.

        Returns:
            The transaction information as a ``Transaction`` object.
        """
        return Transaction.parse_obj(
            await self.eth.get_transaction(transaction_hash.to_web3())
        )

    async def eth_get_raw_transaction(
        self, transaction_hash: Hash32
    ) -> HexBytes:
        """Get the raw bytes of a specific transaction by its hash.

        Args:
            transaction_hash: The hash of the transaction to retrieve.

        Returns:
            The raw bytes of the transaction as a ``HexBytes`` object.
        """
        return HexBytes(
            (
                await self.eth.get_raw_transaction(transaction_hash.to_web3())
            ).hex()
        )

    async def eth_get_transaction_by_block(
        self, block_identifier: BlockIdentifier, index: int
    ) -> Transaction:
        """Get information about a specific transaction by its block and index.

        Args:
            block_identifier: The block identifier containing the transaction.
            index: The index of the transaction within the block.

        Returns:
            The transaction information as a ``Transaction`` object.
        """
        block_id = convert.block_id_transfer(block_identifier)
        return Transaction.parse_obj(
            await self.eth.get_transaction_by_block(block_id, index)
        )

    async def eth_get_raw_transaction_by_block(
        self, block_identifier: BlockIdentifier, index: int
    ) -> HexBytes:
        """Get the raw bytes of a specific transaction by its block and index.

        Args:
            block_identifier: The block identifier containing the transaction.
            index: The index of the transaction within the block.

        Returns:
            The raw bytes of the transaction as a ``HexBytes`` object.
        """
        block_id = convert.block_id_transfer(block_identifier)
        return HexBytes(
            (
                await self.eth.get_raw_transaction_by_block(block_id, index)
            ).hex()
        )

    async def eth_get_transaction_cnt_by_block(
        self, block_identifier: BlockIdentifier
    ) -> int:
        """Get the number of transactions in a specific block.

        Args:
            block_identifier: The block identifier.

        Returns:
            The number of transactions in the block.
        """
        block_id = convert.block_id_transfer(block_identifier)
        return await self.eth.get_block_transaction_count(block_id)

    async def eth_send_transaction(self, transaction: TxParams) -> HexBytes:
        return HexBytes(
            await self.eth.send_transaction(transaction.to_web3())
        )

    async def eth_get_balance(
        self,
        account: Address | ENS,
        block_identifier: BlockIdentifier = "latest"
    ) -> Wei:
        """Get the balance of an Ethereum account.

        Args:
            account: The address or ENS name of the account.
            block_identifier: The block identifier. Defaults to "latest".

        Returns:
            The balance of the account as a ``Wei`` object.
        """
        block_id = convert.block_id_transfer(block_identifier)
        addr = convert.address_transfer(account)
        return Wei(await self.eth.get_balance(addr, block_id))

    async def eth_get_code(
        self,
        account: Address | ENS,
        block_identifier: BlockIdentifier = "latest"
    ) -> HexBytes:
        """Get the bytecode of a contract deployed on the Ethereum network.

        Args:
            account: The address or ENS name of the contract.
            block_identifier: The block identifier. Defaults to "latest".

        Returns:
            The bytecode of the contract as a ``HexBytes`` object.
        """
        block_id = convert.block_id_transfer(block_identifier)
        addr = convert.address_transfer(account)
        return HexBytes((await self.eth.get_code(addr, block_id)).hex())

    async def eth_get_account_nonce(
        self,
        account: Address | ENS,
        block_identifier: BlockIdentifier = "latest"
    ) -> Nonce:
        """Get the nonce of an Ethereum account.

        Args:
            account: The address or ENS name of the account.
            block_identifier: The block identifier. Defaults to "latest".

        Returns:
            The nonce of the account as a ``Nonce`` object.
        """
        block_id = convert.block_id_transfer(block_identifier)
        addr = convert.address_transfer(account)
        return await self.eth.get_transaction_count(addr, block_id)

    async def eth_get_block(
        self,
        block_identifier: BlockIdentifier,
        full_transactions: bool = False
    ) -> Block:
        """Get information about a specific block.

        Args:
            block_identifier: The block identifier.
            full_transactions: Whether to include full transaction details.
                Defaults to ``False``.

        Returns:
            Block: The block information as a ``Block`` object.
        """
        block_id = convert.block_id_transfer(block_identifier)
        return Block.parse_obj(
            await self.eth.get_block(block_id, full_transactions)
        )

    async def eth_get_storage_at(
        self,
        account: Address | ENS,
        position: int,
        block_identifier: BlockIdentifier = "latest",
    ) -> HexBytes:
        """Get the value stored at a specific position in an Ethereum account's
        storage.
    
        Args:
            account: The address or ENS name of the account.
            position: The position in storage to retrieve.
            block_identifier: The block identifier. Defaults to "latest".
        
        Returns:
            The value stored at the specified position as a ``HexBytes``
            object.
        """
        block_id = convert.block_id_transfer(block_identifier)
        addr = convert.address_transfer(account)
        return HexBytes(
            await self.eth.get_storage_at(addr, position, block_id)
        )

    async def eth_send_raw_transaction(
        self, transaction: HexBytes
     ) -> Hash32:
        """Send a raw transaction to the Ethereum network.

        Args:
            transaction: The raw transaction as a ``HexBytes`` object.

        Returns:
            The hash of the sent transaction as a ``Hash32`` object.
        """
        return Hash32(
            await self.eth.send_raw_transaction(transaction.value)
        )

    async def eth_wait_for_transaction_receipt(
        self,
        transaction_hash: Hash32,
        timeout: float = 120,
        poll_latency: float = 0.1
    ) -> Receipt:
        """Wait for a transaction to be included in a block and return the
        receipt.

        Args:
            transaction_hash: The hash of the transaction to wait for.
            timeout: The maximum time to wait in seconds. Defaults to ``120``.
            poll_latency: The time between polling attempts in seconds.
                Defaults to ``0.1``.

        Returns:
            The transaction receipt as a ``Receipt`` object.
        """
        return Receipt.parse_obj(
            await self.eth.wait_for_transaction_receipt(
                transaction_hash.to_web3(), timeout, poll_latency
            )
        )

    async def eth_get_transaction_receipt(
        self, transaction_hash: Hash32
    ) -> Receipt:
        """Get the receipt of a specific transaction.

        Args:
            transaction_hash: The hash of the transaction to retrieve the
            receipt for.

        Returns:
            The transaction receipt as a ``Receipt`` object.
        """
        return Receipt.parse_obj(
            await self.eth.get_transaction_receipt(
                transaction_hash.to_web3()
            )
        )

    async def eth_sign(
        self, account: Address | ENS, data: HexBytes
    ) -> HexBytes:
        """Sign data with an Ethereum account.
    
        Args:
            account: The address or ENS name of the account to sign with.
            data: The data to sign as a ``HexBytes`` object.
        
        Returns:
            HexBytes: The signature as a ``HexBytes`` object.
        """
        addr = convert.address_transfer(account)
        return HexBytes(await self.eth.sign(addr, data.value))

    async def eth_filter(
        self, filter: Literal["latest", "pending"] | FilterParams
    ) -> AsyncFilter:
        """Create a filter to watch for specific events on the Ethereum
        network.

        Args:
            filter: The filter specification.

        Returns:
            The created filter as an ``AsyncFilter`` object.
        """
        if isinstance(filter, FilterParams):
            return await self.eth.filter(filter.to_web3())
        else:
            return await self.eth.filter(filter)

    async def eth_uninstall_filter(self, filter: AsyncFilter) -> bool:
        """Uninstall a previously installed filter.

        Args:
            filter: The filter to uninstall.

        Returns:
            ``True`` if the filter was successfully uninstalled, ``False``
            otherwise.
        """
        assert filter.filter_id is not None
        return await self.eth.uninstall_filter(filter.filter_id)



from typing import (
    Any,
)

from eth_typing import (
    Address as Web3Address,
)
from pydantic import (
    validator,
)
from web3.types import (
    ENS,
    BlockIdentifier as Web3BlockIdentifier,
)

from ethhelper.datatypes.base import (
    Address,
    BlockIdentifier,
    Hash32,
)

Validator = Any


def parse_hex_or_strint(v: Any) -> Any:
    if not isinstance(v, str):
        return v
    return int(v, 0)


def block_id_transfer(
    block_identifier: BlockIdentifier
) -> Web3BlockIdentifier:
    if isinstance(block_identifier, Hash32):
        return block_identifier.to_web3()
    return block_identifier


def address_transfer(address: Address | ENS) -> Web3Address | ENS:
    if isinstance(address, Address):
        return address.to_web3()
    return address


def int_validator(*fields: str) -> Validator:
    return validator(*fields, pre=True, allow_reuse=True)(
        parse_hex_or_strint
    )



from logging import (
    Logger,
)

from web3.net import (
    AsyncNet,
)

from .base import (
    GethHttpWeb3,
)


class GethNetHttp(GethHttpWeb3):
    """Class for interacting with a Geth node using HTTP with the Geth JSON-RPC
    API.

    This class extends the GethHttpWeb3 class and provides additional methods
    specifically for interacting with the network using the Geth JSON-RPC API.

    The ``url`` is used to indicate the path of the HTTP service of the Geth
    node, usually in the form of ``http://host:port/``. The use of third-party
    nodes may be out of the ordinary.

    The ``logger`` is used to assign a logger of the Python logging module to
    this class. Explicitly assigning a logger can be used to control the output
    location of the logger, which is convenient for debugging.
    """
    def __init__(self, url: str, logger: Logger) -> None:
        super().__init__(url, logger)
        self.net: AsyncNet = self.w3.net
        """Asynchronous Net interface for Web3. Used to simplify the access
        path.
        """

    async def net_listening(self) -> bool:
        """Check if the Geth node is currently listening for connections.

        Returns:
            ``True`` if the node is listening, ``False`` otherwise.
        """
        return await self.net.listening

    async def net_peer_count(self) -> int:
        """Get the number of peers currently connected to the Geth node.

        Returns:
            The number of peers.
        """
        return await self.net.peer_count

    async def net_version(self) -> str:
        """Get the version number of the Geth node.

        Returns:
            str: The version as a string.
        """
        return await self.net.version



from typing import (
    Any,
    TypeVar,
)

import orjson

from ethhelper.datatypes.base import (
    HexBytes,
    IntStr,
)

T = TypeVar("T", IntStr, HexBytes)


def encode_my_class(v: Any) -> Any:
    if isinstance(v, IntStr):
        return v.to_int_or_str()
    elif isinstance(v, HexBytes):
        return str(v)
    raise TypeError


def orjson_dumps(v: Any, **kwargs: Any) -> str:
    return orjson.dumps(
        v, default=encode_my_class, option=orjson.OPT_NON_STR_KEYS
    ).decode()



from abc import (
    abstractmethod,
)
import asyncio
from asyncio import (
    Event,
)
import logging
from logging import (
    Logger,
)

from ethhelper.datatypes.eth import (
    Block,
)
from ethhelper.datatypes.geth import (
    GethIsDead,
    GethSuccessResponse,
    GethWSResponse,
    NoSubscribeToken,
)

from .base import (
    GethSubscriber,
)


class GethNewBlockSubscriber(GethSubscriber):
    def __init__(self, url: str, logger: Logger | None = None) -> None:
        if logger is None:
            logger = logging.getLogger("GethNewBlockSubsriber")
        super().__init__(url, logger)

    async def subscribe_new_block(self) -> None:
        """Subscribe to new block notifications on the Geth node."""
        self.subscribe_id = await self.subscribe("newHeads")

    async def after_connection(self) -> None:
        """A method that is called after the connection to the Geth node has
        been established.

        This method is overridden to wait for the Geth node to finish syncing
        and then subscribe to new block notifications.
        """
        self.wait_syncing = True
        self.syncing = Event()
        self.wait_first = True
        while self.wait_syncing:
            await self.send("eth_syncing", [])
            self.logger.info("Waiting for the result of syncing")
            await self.syncing.wait()
            if self.wait_syncing:
                self.logger.warning("Geth node is syncing...")
                await asyncio.sleep(5)
        self.logger.info("Geth is synced. Continue.")
        await self.subscribe_new_block()
        

    async def handle(self, data: GethWSResponse | GethSuccessResponse) -> None:
        """A method that is called when a message is received from the Geth
        node.

        This method is overridden to handle new block notifications from the
        Geth node.

        Args:
            data: The message received from the Geth node.
        """
        if self.wait_syncing:
            if not isinstance(data, GethSuccessResponse):
                raise GethIsDead
            if isinstance(data.result, bool) and not data.result:
                self.wait_syncing = False
            self.syncing.set()
            return
        if self.wait_first:
            if not isinstance(data, GethSuccessResponse):
                raise NoSubscribeToken
            if not isinstance(data.result, str):
                raise NoSubscribeToken
            self.subscribe_token = data.result
            self.wait_first = False
            return
        if not isinstance(data, GethWSResponse):
            await self.on_other(data)
            return

        await self.on_block(Block.parse_obj(data.params.result))

    @abstractmethod
    async def on_block(self, block: Block) -> None:
        """A method that is called when a new block notification is received
        from the Geth node.

        This method must be overridden to handle new block notifications from
        the Geth node.

        Args:
            block: The new block that was received from the Geth node.
        """
        raise NotImplementedError

    async def on_other(self, data: GethSuccessResponse) -> None:
        """A method that is called when a non-new-block message is received
        from the Geth node.

        This method can be overridden to handle non-new-block messages from the
        Geth node.

        Args:
            data: The message received from the Geth node.
        """
        pass



from typing import (
    Any,
)

from eth_typing import (
    ChecksumAddress,
)
import orjson
from pydantic import (
    BaseModel,
)
from web3.types import (
    Nonce,
)

from ethhelper.utils import (
    convert,
    json,
)

from .base import (
    Address,
    Gas,
    Wei,
)
from .eth import (
    Transaction,
)


class TxpoolStatus(BaseModel):
    """A class that represents the status of transactions in the Ethereum
    transaction pool.
    """
    pending: int
    """The number of pending transactions in the pool."""
    queued: int
    """The number of queued transactions in the pool."""
    # vaildators
    int_val = convert.int_validator("pending", "queued")

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        frozen = True
        json_loads = orjson.loads
        json_dumps = json.orjson_dumps


class TxpoolSnapshot(BaseModel):
    """A class that represents a snapshot of the Ethereum transaction pool.
    """

    contract_creation: bool = False
    """Whether the transaction is a contract creation. Defaults to ``False``.
    """
    to_address: Address
    """The address of the recipient of the transaction."""
    fee: Wei
    """The transaction fee in wei."""
    gas: Gas
    """The amount of gas used by the transaction."""
    gas_fee: Wei
    """The gas fee for the transaction in wei."""
    # vaildators
    int_val = convert.int_validator("fee", "gas", "gas_fee")

    @classmethod
    def validate(cls, value: Any) -> "TxpoolSnapshot":
        """Validates the given value and creates a new instance of the
        ``TxpoolSnapshot`` class.

        Args:
            value: The value to be validated and converted.

        Returns:
            The converted ``TxpoolSnapshot`` instance.

        """
        if not isinstance(value, str):
            return super(TxpoolSnapshot, cls).validate(value)
        ss = value.split(": ")
        contract_creation = ss[0] == "contract creation"
        if contract_creation:
            ss[0] = "0x0000000000000000000000000000000000000000"
        to_address = Address(ss[0])
        ss = ss[1].split(" + ")
        fee = Wei(ss[0][:-4])
        ss = ss[1].split(" × ")
        gas = Gas(int(ss[0][:-4]))
        gas_fee = Wei(ss[1][:-4])
        return cls(
            contract_creation=contract_creation,
            to_address=to_address,
            fee=fee,
            gas=gas,
            gas_fee=gas_fee
        )

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        frozen = True
        json_loads = orjson.loads
        json_dumps = json.orjson_dumps


class TxpoolInspect(BaseModel):
    """A class that represents the contents of the Ethereum transaction pool.
    """
    pending: dict[ChecksumAddress, dict[Nonce, TxpoolSnapshot]]
    """A dictionary that maps from account address to nonce to snapshot of the
    corresponding transaction in the pending pool.
    """
    queued: dict[ChecksumAddress, dict[Nonce, TxpoolSnapshot]]
    """A dictionary that maps from account address to nonce to snapshot of the
    corresponding transaction in the queued pool."""

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        frozen = True
        json_loads = orjson.loads
        json_dumps = json.orjson_dumps


class TxpoolContent(BaseModel):
    """A class representing the content of the transaction pool on a node.
    """
    pending: dict[ChecksumAddress, dict[Nonce, Transaction]]
    """A dictionary containing information about all pending transactions in
    the pool.
    
    The keys are the transaction sender's address and the values are
    another dictionary, where the keys are the nonces of the transactions and
    the values are the corresponding transactions.
    """
    queued: dict[ChecksumAddress, dict[Nonce, Transaction]]
    """A dictionary containing information about all queued transactions in the
    pool.
    
    The keys are the transaction sender's address and the values are
    another dictionary, where the keys are the nonces of the transactions and
    the values are the corresponding transactions.
    """

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        frozen = True
        json_loads = orjson.loads
        json_dumps = json.orjson_dumps


class TxpoolContentFrom(BaseModel):
    """A class that represents the transactions in the transaction pool
    (txpool) from a specific address.
    """
    pending: dict[Nonce, Transaction]
    """A dictionary of pending transactions from the specified address. The
    dictionary is indexed by the nonce of each transaction."""
    queued: dict[Nonce, Transaction]
    """A dictionary of queued transactions from the specified address. The
    dictionary is indexed by the nonce of each transaction."""

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        frozen = True
        json_loads = orjson.loads
        json_dumps = json.orjson_dumps



from decimal import (
    Decimal,
)
from typing import (
    Any,
    NewType,
)

from eth_typing import (
    Address as Web3Address,
    BlockNumber,
    Hash32 as EthHash32,
)
from hexbytes import (
    HexBytes as Web3HexBytes,
)
from web3.types import (
    BlockParams,
    Wei as Web3Wei,
)


class IntStr:
    """A class that represents an integer value that can be initialized
    from a string or another ``IntStr`` instance.
    
    This class is designed to work with the ``orjson`` library, and provides a
    way to represent integer values that may be too large to be encoded using
    64 bits.
    
    By using an ``IntStr`` instance to represent such values, you can ensure
    that the integers are passed around as Python integers but still allow them
    to be encoded as strings if necessary.

    The ``value`` is the value to be initialized. If a string is provided, it
    will be converted to an integer.
    """
    def __init__(self, value: "str | int | IntStr") -> None:
        self.value: int
        """An integer value."""
        if isinstance(value, IntStr):
            self.value = value.value
        elif isinstance(value, int):
            self.value = value
        else:
            assert isinstance(value, str)
            self.value = int(value, 0)

    def to_int_or_str(self) -> int | str:
        """Returns the integer value as a string if it has 64 or more bits,
        otherwise it returns the integer value itself.

        Returns:
            The integer value as a string if it has 64 or more bits, otherwise
            the integer value itself.
        """
        if self.value.bit_length() >= 64:
            return str(self.value)
        return self.value

    def __hash__(self) -> int:
        return hash(self.value)

    def __lt__(self, __o: Any) -> bool:
        o = IntStr(__o)
        return self.value < o.value

    def __le__(self, __o: Any) -> bool:
        o = IntStr(__o)
        return self.value <= o.value

    def __eq__(self, __o: Any) -> bool:
        o = IntStr(__o)
        return self.value == o.value

    def __ne__(self, __o: Any) -> bool:
        o = IntStr(__o)
        return self.value != o.value

    def __ge__(self, __o: Any) -> bool:
        o = IntStr(__o)
        return self.value >= o.value

    def __gt__(self, __o: Any) -> bool:
        o = IntStr(__o)
        return self.value > o.value

    def __str__(self) -> str:
        return str(self.value)

    @classmethod
    def __get_validators__(cls) -> Any:
        yield cls.validate

    @classmethod
    def validate(cls, value: Any) -> Any:
        """Validates and returns the value as an ``IntStr`` instance.

        Args:
            value: The value to be validated and returned.

        Returns:
            An ``IntStr`` instance that represents the validated value.

        Raises:
            TypeError: If the value is not an ``IntStr``, int, or str.
        """
        if isinstance(value, IntStr):
            return value
        if isinstance(value, int) or isinstance(value, str):
            return cls(value)
        return value


class HexBytes:
    """A class that represents a byte string that can be initialized from a
    string, bytes, ``hexbytes.HexBytes`` instance or another ``HexBytes``
    instance.
    """
    def __init__(self, value: "str | bytes | HexBytes") -> None:
        self.value: bytes
        if isinstance(value, HexBytes):
            self.value = value.value
        elif isinstance(value, Web3HexBytes):
            self.value = bytes(value)
        elif isinstance(value, bytes):
            self.value = value
        else:
            assert isinstance(value, str)
            if not value.startswith("0x"):
                raise ValueError("HexBytes should start with 0x.")
            hex_str = value[2:]
            if len(hex_str) % 2 == 1:
                hex_str = f"0{hex_str}"
            self.value = bytes.fromhex(hex_str)

    def __hash__(self) -> int:
        return hash(self.value)

    def __lt__(self, __o: Any) -> bool:
        o = HexBytes(__o)
        return self.value < o.value

    def __le__(self, __o: Any) -> bool:
        o = HexBytes(__o)
        return self.value <= o.value

    def __eq__(self, __o: Any) -> bool:
        o = HexBytes(__o)
        return self.value == o.value

    def __ne__(self, __o: Any) -> bool:
        o = HexBytes(__o)
        return self.value != o.value

    def __ge__(self, __o: Any) -> bool:
        o = HexBytes(__o)
        return self.value >= o.value

    def __gt__(self, __o: Any) -> bool:
        o = HexBytes(__o)
        return self.value > o.value

    def __str__(self) -> str:
        return f"0x{self.value.hex()}"

    @classmethod
    def __get_validators__(cls) -> Any:
        yield cls.validate

    @classmethod
    def validate(cls, value: Any) -> Any:
        """Validates and converts a value to a ``HexBytes`` instance.

        Args:
            value: The value to be validated and converted.

        Returns:
            A ``HexBytes`` instance representing the input value.

        Raises:
            TypeError: If the input value is not a ``HexBytes``, bytes, or
                string starting with "0x".
        """
        if isinstance(value, HexBytes):
            return value
        if isinstance(value, bytes) or \
                (isinstance(value, str) and value.startswith("0x")):
            return cls(value)
        return value


class Hash32(HexBytes):
    """A subclass of ``HexBytes`` that represents a 32-byte hash value."""
    def to_web3(self) -> EthHash32:
        """Returns an ``eth_typing.Hash32`` instance that represents the hash
        value.

        Returns:
            An ``eth_typing.Hash32`` instance that represents the hash value.
        """
        return EthHash32(self.value)


class Address(HexBytes):
    """A subclass of HexBytes that represents an Ethereum address (20 bytes).
    """
    def to_web3(self) -> Web3Address:
        """Returns a ``eth_typing.Address`` instance that represents the
        address.

        Returns:
            A ``eth_typing.Address`` instance that represents the address.
        """
        return Web3Address(self.value)


BlockIdentifier = BlockParams | BlockNumber | Hash32
"""A union type that can represent a block number, a block hash, or the strings
"latest", "earliest", or "pending".
"""
Gas = NewType("Gas", int)
"""A new type that represents a gas value (an integer)."""


class Wei(IntStr):
    """A subclass of ``IntStr`` that represents a value in wei (the smallest
    unit of ether in Ethereum).
    """
    def to_web3(self) -> Web3Wei:
        """Returns a ``web3.types.Wei`` instance that represents the value in
        wei.

        Returns:
            A ``web3.types.Wei`` instance that represents the value in wei.
        """
        return Web3Wei(self.value)
    
    @classmethod
    def from_gwei(cls, value: int) -> "Wei":
        """Class method that converts a value in gigawei to wei.

        Args:
            value: An integer value in gigawei.

        Returns:
            A ``Wei`` instance that represents the converted value in wei.
        """
        return cls(value * 1_000_000_000)

    @classmethod
    def from_eth(cls, value: int | float | Decimal) -> "Wei":
        """Class method that converts a value in ether to wei.

        Args:
            value: A value in ether.

        Returns:
            Wei: A ``Wei`` instance that represents the converted value in wei.
        """
        return cls(int(value * 1_000_000_000_000_000_000))




from abc import (
    ABCMeta,
    abstractmethod,
)
import asyncio
from asyncio import (
    CancelledError,
    Task,
)
from logging import (
    Logger,
)
import traceback
from typing import (
    Any,
)

from pydantic import (
    ValidationError,
)
from websockets import (
    client,
)

from ethhelper.datatypes.geth import (
    GethError,
    GethErrorResponse,
    GethRequest,
    GethSuccessResponse,
    GethWSResponse,
)


class GethSubscriber(metaclass=ABCMeta):
    """Abstract base class for implementing a subscriber to a Geth node using
    websockets with the Geth JSON-RPC API.

    This class defines the basic structure for implementing a subscriber to a
    Geth node using websockets. It provides a method for binding to the node,
    sending requests, handling responses, and closing the connection.

    The ``url`` is used to indicate the path of the WS service of the Geth
    node, usually in the form of ``http://host:port/``. The use of third-party
    nodes may be out of the ordinary.

    The ``logger`` is used to assign a logger of the Python logging module to
    this class. Explicitly assigning a logger can be used to control the output
    location of the logger, which is convenient for debugging.
    """
    def __init__(self, url: str, logger: Logger) -> None:
        self.url = url
        self.logger = logger
        self.run_task: Task[None] | None = None
        self.closed = False

    async def bind(self) -> Task[None]:
        """Bind the subscriber to the Geth node and start listening for
        messages.

        Returns:
            A task that will run the subscriber until it is closed.
        """
        if self.closed:
            self.logger.error("Cant rebind closed subsriber!")
            raise ValueError("Cant rebind closed subsriber!")
        self.run_task = asyncio.create_task(self.run())
        return self.run_task

    async def run(self) -> None:
        """The main loop that listens for messages from the Geth node."""
        while not self.closed:
            self.id = 0
            try:
                async with client.connect(self.url) as self.ws:
                    self.recv_loop = asyncio.create_task(self._recieve_loop())
                    await self.after_connection()
                    await self.recv_loop
                    exce = self.recv_loop.exception()
                    if exce is not None:
                        self.logger.warning(
                            "Error when recv in Websocket. Retry is 5s."
                        )
                        self.logger.debug(
                            f"Details: {traceback.format_exception(exce)}"
                        )
                        await asyncio.sleep(5)
            except Exception:
                self.logger.warning(
                    "Websocket connection is dead. Retry is 5s."
                )
                self.logger.debug(f"Details: {traceback.format_exc()}")
                await asyncio.sleep(5)

    async def send(self, method: str, params: list[Any]) -> int:
        """Send a request to the Geth node over the websocket connection.

        Args:
            method: The JSON-RPC method to call.
            params: The parameters to send with the request.

        Returns:
            The ID of the request, which can be used to match responses with
            requests.
        """
        self.id += 1
        data = GethRequest(id=self.id, method=method, params=params).json()
        self.logger.debug(f"SEND {data}")
        await self.ws.send(data)
        return self.id

    async def _recieve_loop(self) -> None:
        """The loop that listens for messages from the Geth node."""
        async for data in self.ws:
            if isinstance(data, bytes):
                data = data.decode()
            self.logger.debug(f"RECV {data}")
            response: GethWSResponse | GethSuccessResponse | GethErrorResponse
            try:
                response = GethWSResponse.parse_raw(data)
            except ValidationError:
                try:
                    response = GethSuccessResponse.parse_raw(data)
                except ValidationError:
                    response = GethErrorResponse.parse_raw(data)
                    raise GethError(error=response.error)
            await self.handle(response)

    async def subscribe(self, param: str) -> int:
        """Subscribe to a specific event or method on the Geth node.
        
        Args:
            param: The name of the event or method to subscribe to.

        Returns:
            The ID of the subscription, which can be used to unsubscribe later.
        """
        return await self.send("eth_subscribe", [param])

    @abstractmethod
    async def after_connection(self) -> None:
        """A method that is called after the connection to the Geth node has
        been established.

        This method can be overridden to perform any necessary setup or
        initialization after the connection to the Geth node has been
        established.
        """
        raise NotImplementedError

    @abstractmethod
    async def handle(self, data: GethWSResponse | GethSuccessResponse) -> None:
        """A method that is called when a message is received from the Geth
        node.

        This method must be overridden to handle the messages received from the
        Geth node.

        Args:
            data: The message received from the Geth node.
        """
        raise NotImplementedError

    async def close(self) -> None:
        """Close the connection to the Geth node."""
        if self.closed:
            return
        self.closed = True
        if self.run_task is None:
            return
        await self.ws.close()
        self.recv_loop.cancel()
        self.run_task.cancel()
        try:
            await self.run_task
        except CancelledError:
            pass



from eth_typing import (
    HexAddress,
)

from ethhelper.datatypes.txpool import (
    TxpoolContent,
    TxpoolContentFrom,
    TxpoolInspect,
    TxpoolStatus,
)

from .base import (
    GethHttpCustomized,
)


class GethTxpoolHttp(GethHttpCustomized):
    """A class that provides an interface to interact with the Geth txpool
    using HTTP.

    The ``GethTxpoolHttp`` class inherits from the ``GethHttpCustomized`` class
    and provides methods to interact with the Geth txpool using HTTP.

    The ``url`` is used to indicate the path of the HTTP service of the Geth
    node, usually in the form of ``http://host:port/``. The use of third-party
    nodes may be out of the ordinary.

    The ``logger`` is used to assign a logger of the Python logging module to
    this class. Explicitly assigning a logger can be used to control the output
    location of the logger, which is convenient for debugging.
    """
    async def txpool_status(self) -> TxpoolStatus:
        """Returns an object representing the status of the Geth txpool.

        This function sends a ``txpool_status`` request to the Geth node and
        returns the response as a ``TxpoolStatus`` object.

        Returns:
            A ``TxpoolStatus`` object representing the status of the Geth
            txpool.

        Raises:
            ethhelper.types.GethError: Raised when response is a Geth error.
            ethhelper.types.IdNotMatch: Raised when received response id not
                match request id.
        """
        result = await self.send("txpool_status")
        return TxpoolStatus.parse_obj(result)

    async def txpool_inspect(self) -> TxpoolInspect:
        """Returns an object representing the contents of the Geth txpool.

        This function sends a ``txpool_inspect`` request to the Geth node and
        returns the response as a ``TxpoolInspect`` object.

        Returns:
            A ``TxpoolInspect`` object representing the contents of the Geth
            txpool.

        Raises:
            ethhelper.types.GethError: Raised when response is a Geth error.
            ethhelper.types.IdNotMatch: Raised when received response id not
                match request id.
        """
        result = await self.send("txpool_inspect")
        return TxpoolInspect.parse_obj(result)

    async def txpool_content(self) -> TxpoolContent:
        """Returns an object representing the contents of the Geth txpool.

        This function sends a ``txpool_content`` request to the Geth node and
        returns the response as a ``TxpoolContent`` object.

        Returns:
            A ``TxpoolContent`` object representing the contents of the Geth
            txpool.

        Raises:
            ethhelper.types.GethError: Raised when response is a Geth error.
            ethhelper.types.IdNotMatch: Raised when received response id not
                match request id.
        """
        result = await self.send("txpool_content")
        return TxpoolContent.parse_obj(result)

    async def txpool_content_from(
            self, address: HexAddress) -> TxpoolContentFrom:
        """Returns an object representing the transactions from a specific
        address in the Geth txpool.

        This function sends a ``txpool_contentFrom`` request to the Geth node
        with the specified address and returns the response as a
        ``TxpoolContentFrom`` object.

        Args:
            address: The address to get the transactions from.

        Returns:
            A ``TxpoolContentFrom`` object representing the transactions from
            the specified address in the Geth txpool.

        Raises:
            ethhelper.types.GethError: Raised when response is a Geth error.
            ethhelper.types.IdNotMatch: Raised when received response id not
                match request id.
        """
        result = await self.send("txpool_contentFrom", [address])
        return TxpoolContentFrom.parse_obj(result)



import abc
from abc import (
    ABCMeta,
)
from logging import (
    Logger,
)
import traceback
from typing import (
    Any,
)

from httpx import (
    AsyncClient,
)
import orjson
from pydantic import (
    ValidationError,
)
from web3 import (
    AsyncHTTPProvider,
    AsyncWeb3,
)

from ethhelper.datatypes.geth import (
    GethError,
    GethErrorResponse,
    GethRequest,
    GethResponse,
    GethSuccessResponse,
    IdNotMatch,
)
from ethhelper.utils import (
    json,
)


class GethHttpAbstract(metaclass=ABCMeta):
    """A basic abstraction over Geth's HTTP interface wrapper.

    The ``url`` is used to indicate the path of the HTTP service of the Geth
    node, usually in the form of ``http://host:port/``. The use of third-party
    nodes may be out of the ordinary.

    The ``logger`` is used to assign a logger of the Python logging module to
    this class. Explicitly assigning a logger can be used to control the output
    location of the logger, which is convenient for debugging.
    """

    def __init__(self, url: str, logger: Logger) -> None:
        self.url: str = url
        """The url giving from the constructor of this class.

            >>> from ethhelper import GethHttpConnector
            >>> connector = GethHttpConnector("http://localhost:8545/")
            >>> connector.url
            'http://localhost:8545/'
        """
        self.logger: Logger = logger
        """The logger giving from the constructor of this class or default by
        ``logging.getLogger("GethHttpConnector")``
        """

    @abc.abstractmethod
    async def is_connected(self) -> bool:
        """Check connectivity.
        
        Subclasses must implement methods to detect connectivity.

        Returns:
            A bool value for connectivity is good or not.
        """
        raise NotImplementedError()


class GethHttpCustomized(GethHttpAbstract):
    """An interface wrapper for direct HTTP access to Geth.

    The `GethHttpCustomized` class provides an interface wrapper that
    allows for direct HTTP access to a Geth node.

    The ``url`` is used to indicate the path of the HTTP service of the Geth
    node, usually in the form of ``http://host:port/``. The use of third-party
    nodes may be out of the ordinary.

    The ``logger`` is used to assign a logger of the Python logging module to
    this class. Explicitly assigning a logger can be used to control the output
    location of the logger, which is convenient for debugging.
    """

    def __init__(self, url: str, logger: Logger) -> None:
        super().__init__(url, logger)
        self.id: int = 0
        """The id used when sending requests to Geth, an integer starting from
        ``1`` and incrementing for each request. Reset when it exceeds
        ``100000000``.
        """

    def parse_response(self, raw_res: str) -> GethResponse:
        """Parse the content of the Geth node response.

        This function will attempt to parse the input into a
        ``GethSuccessResponse``, and failing that, into a
        ``GethErrorResponse``.

        Args:
            raw_res: the content of the HTTP response of the Geth node in
                string form.
        
        Returns:
            An object of either GethSuccessResponse or GethErrorResponse,
            respectively indicates whether this is a normalized success
            response or an error response.
        
        Raises:
            pydantic.ValidationError: Raised when input data cannot be parsed
                into either GethSuccessResponse or GethErrorResponse.
        """
        try:
            response = GethSuccessResponse.parse_raw(raw_res)
        except ValidationError:
            response = GethErrorResponse.parse_raw(raw_res)
        self.logger.debug(f"RECV {response}")
        return response

    def parse_multiple_responses(
        self, raw_res: str
    ) -> tuple[list[GethSuccessResponse], list[GethErrorResponse]]:
        """Parse the content of the Geth node responses.

        Geth provides an interface that allows multiple Geth requests to be
        served in a single HTTP request. Likewise, the response will be
        multiple. This function is used to parse multiple Geth responses from
        a single HTTP response.

        This function will attempt to parse the input into a list. Afterwards,
        this function assumes that the elements of this list are in the form of
        ``GethSuccessResponse` or ``GethErrorResponse``, and use pydantic to
        parse these elements.

        Args:
            raw_res: the content of the HTTP response of the Geth node in
                string form.

        Returns:
            A tuple with two lists. One is a list of GethSuccessResponse and
            the other one is a list of GethErrorResponse. They respectively
            indicate whether these are normalized success responses or error
            responses.

        Raises:
            orjson.JSONDecodeError: Raised when the raw content cannot be json
                decoded.
            pydantic.ValidationError: Raised when any element in json decoded
                list cannot be parsed into either GethSuccessResponse or
                GethErrorResponse.
        """
        raw_res_list = orjson.loads(raw_res)
        success: list[GethSuccessResponse] = []
        errors: list[GethErrorResponse] = []
        for res in raw_res_list:
            try:
                response = GethSuccessResponse.parse_obj(res)
                success.append(response)
            except ValidationError:
                response = GethErrorResponse.parse_obj(res)
                errors.append(response)
        self.logger.debug(f"RECV MULTIPLE {success} {errors}")
        return success, errors

    async def send_raw(self, raw: str) -> str:
        """Send json text to Geth node and return text of the response.

        Args:
            raw: The json text will be sent.

        Returns:
            A string of the json content of the response in text.

        Raises:
            httpx.RequestError: Raised when an HTTP request fails.
        """
        self.logger.debug(f"SEND RAW {raw}")
        async with AsyncClient() as client:
            res = await client.post(
                f"{self.url}",
                content=raw,
                headers={"Content-Type": "application/json"}
            )
            self.logger.debug(f"RECV RAW {res.text}")
            return res.text

    async def send(self, method: str, params: list[Any] | None = None) -> Any:
        """Send a Geth request to Geth node and return the data of Geth
        response.

        HTTP communication with Geth follows the JSON-RPC specification. It
        stipulates that each request needs to provide a ``method`` and a series
        of ``parameters``. This function is the encapsulation of this behavior.
        An ``id`` is automatically generated inside the function. The function
        will form a ``GethRequest`` object with the ``id`` and the input
        ``method`` and ``params``, encode it into json text by pydantic and
        send it to the Geth node.

        Args:
            method: The method name of the Geth HTTP interface to call.
            params: A series of parameters used by Geth to make the request.

        Returns:
            A basic type of object that represents the result returned by Geth
            after executing the request.

        Raises:
            httpx.RequestError: Raised when an HTTP request fails.
            ethhelper.types.GethError: Raised when response is a Geth error.
            ethhelper.types.IdNotMatch: Raised when received response id not
                match request id.
        """
        if params is None:
            params = []
        self.logger.debug(f"SEND {method} {params}")
        if self.id >= 100000000:
            self.id = 0
        self.id += 1
        id = self.id
        request = GethRequest(id=id, method=method, params=params)
        raw_res = await self.send_raw(request.json())
        response = self.parse_response(raw_res)
        if isinstance(response, GethErrorResponse):
            raise GethError(error=response.error)
        if id != response.id:
            raise IdNotMatch(
                f"Send id {id} but received {response.id}"
            )
        return response.result

    async def send_multiple(
        self, raw_requests: list[tuple[str, list[Any] | None]]
    ) -> tuple[list[GethSuccessResponse], list[GethErrorResponse]]:
        """Send multiple Geth requests to the Geth node and return the data of
        the Geth responses.

        HTTP communication with Geth follows the JSON-RPC specification. It
        stipulates that each request needs to provide a ``method`` and a series
        of ``parameters``. This function is the encapsulation of this behavior.
        Multiple requests can be sent in one HTTP request. 

        Args:
            raw_requests: A list of tuples, where each tuple contains a string
                ``method`` and a list of parameters ``params``. The parameters
                can be ``None`` if there are no parameters for the method.

        Returns:
            A tuple with two lists. One is a list of GethSuccessResponse and
            the other one is a list of GethErrorResponse. They respectively
            indicate whether these are normalized success responses or error
            responses.

        Raises:
            httpx.RequestError: Raised when an HTTP request fails.
            orjson.JSONDecodeError: Raised when the raw content cannot be json
                decoded.
            pydantic.ValidationError: Raised when any element in the json
                decoded list cannot be parsed into either GethSuccessResponse
                or GethErrorResponse.
        """
        self.logger.debug(f"SEND MULTIPLE {raw_requests}")
        requests: list[GethRequest] = []
        for method, params in raw_requests:
            if self.id >= 100000000:
                self.id = 0
            self.id += 1
            if params is None:
                params = []
            req = GethRequest(id=self.id, method=method, params=params)
            requests.append(req)
        raw_res = await self.send_raw(
            json.orjson_dumps([req.dict() for req in requests])
        )
        return self.parse_multiple_responses(raw_res)

    async def is_connected(self) -> bool:
        """Checks the connectivity of the Geth node.

        This function checks the connectivity of the Geth node by sending a
        ``net_version`` request to the node. If the request is successful, it
        indicates that the Geth node is available and returns ``True``. If not,
        it returns ``False``. If an exception occurs during the request, the
        function logs a warning and returns ``False``.

        Returns:
            A bool value indicating whether the Geth node is connected and
            available or not.
        """
        try:
            await self.send("net_version")
            self.logger.info("GethHttpCustomized is good for connection.")
            return True
        except Exception:
            self.logger.warning(
                f"GethHttpCustomized cannot connect to {self.url}."
            )
            self.logger.debug(f"Detail: {traceback.format_exc()}")
            return False


class GethHttpWeb3(GethHttpAbstract):
    """A wrapper based on Web3.py to access Geth HTTP interface.

    The ``url`` is used to indicate the path of the HTTP service of the Geth
    node, usually in the form of ``http://host:port/``. The use of third-party
    nodes may be out of the ordinary.

    The ``logger`` is used to assign a logger of the Python logging module to
    this class. Explicitly assigning a logger can be used to control the output
    location of the logger, which is convenient for debugging.
    """

    def __init__(self, url: str, logger: Logger) -> None:
        super().__init__(url, logger)
        self.w3 = AsyncWeb3(AsyncHTTPProvider(self.url))

    async def is_connected(self) -> bool:
        """Checks the connectivity of the Geth node.

        This function checks the connectivity of the Geth node by calling the
        ``is_connected()`` method of the Web3.py library. If the method returns
        a boolean value, it indicates that the Geth node is available and
        returns ``True``. If not, it returns ``False``. If an exception occurs
        during the request, the function logs a warning and returns ``False``.

        Raises:
            NotImplementedError: If the ``is_connected()`` method of the
                Web3.py library returns a boolean value, it indicates that
                Web3.py is not running in asyncio.

        Returns:
            A bool value indicating whether the Geth node is connected and
            available or not.
        """
        connected = self.w3.is_connected()
        if isinstance(connected, bool):
            self.logger.fatal("GethHttpWeb3 is not running in asyncio!")
            raise NotImplementedError(
                "GethHttpWeb3 is not running in asyncio!"
            )
        return await connected



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['ethhelper',
 'ethhelper.connectors',
 'ethhelper.connectors.http',
 'ethhelper.connectors.ws',
 'ethhelper.datatypes',
 'ethhelper.utils']

package_data = \
{'': ['*']}

install_requires = \
['eth-typing>=3.3.0,<4.0.0',
 'httpx>=0.23.3,<0.24.0',
 'orjson>=3.8.12,<4.0.0',
 'pydantic>=1.10.7,<2.0.0',
 'web3>=6.3.0,<7.0.0',
 'websockets>=10.4,<11.0']

setup_kwargs = {
    'name': 'ethhelper',
    'version': '0.3.6',
    'description': 'Asynchronous Geth node connection encapsulation based on httpx, websockets and web3.py. Geth node and Ethereum type extension based on pydantic.',
    'long_description': "# ETHHelper\n\n[![Build Status](https://img.shields.io/github/actions/workflow/status/XiaoHuiHui233/ETHHelper/publish.yml)](https://github.com/XiaoHuiHui233/ETHHelper/actions)\n[![Documentation Status](https://readthedocs.org/projects/ethhelper/badge/?version=latest)](https://ethhelper.readthedocs.io/en/latest/?badge=latest)\n![Python Version](https://img.shields.io/pypi/pyversions/ethhelper)\n![Wheel Status](https://img.shields.io/pypi/wheel/ethhelper)\n[![Latest Version](https://img.shields.io/github/v/release/XiaoHuiHui233/ETHHelper)](https://github.com/XiaoHuiHui233/ETHHelper/releases)\n[![License](https://img.shields.io/github/license/XiaoHuiHui233/ETHHelper)](https://github.com/XiaoHuiHui233/ETHHelper/blob/main/LICENSE)\n\nAsynchronous Geth node connection encapsulation based on httpx, websockets and web3.py. Geth node and Ethereum type extension based on pydantic.\n\nQuickstart see [this](https://ethhelper.readthedocs.io/en/latest/quickstart.html).\n\n[中文](docs/README_cn.md) | English\n\n## Usage\n\n### pypi\n\nIf you prefer to use pypi to install this package, you can just run the following command:\n\n```bash\npip install ethhelper\n```\n\n### git\n\nThe project is managed by poetry. If you prefer to use git to install this package, you can use poetry to directly add a reference to the project's build package through git.\n\nThe command is as follow:\n\n```bash\npoetry add git+ssh://git@github.com:XiaoHuiHui233/ETHHelper.git\n```\n\n## Build\n\nYou can use poetry's tools to generate a build of this project (pure Python).\n\nThe command is as follow:\n\n```bash\npoetry build\n```\n\n## Author and License\n\nETHHelper was written by [XiaoHuiHui233](https://github.com/XiaoHuiHui233/), licensed under the Apache 2.0.\n",
    'author': 'XiaoHuiHui233',
    'author_email': '1025184872@qq.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/XiaoHuiHui233/ETHHelper',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import typing
from typing import (
    Any,
)

from eth_typing import (
    ChecksumAddress,
)
import orjson
from pydantic import (
    BaseModel,
    Field,
    validator,
)
from web3.types import (
    CallOverrideParams as Web3CallOverrideParams,
    Nonce,
)

from ethhelper.utils import (
    json,
)

from .base import (
    HexBytes,
    Wei,
)


class IdNotMatch(Exception):
    """An exception raised when a JSON-RPC response ID doesn't match the
    request ID.
    """
    pass


class GethIsDead(Exception):
    """An exception raised when the Geth client is not running."""
    pass


class NoSubscribeToken(Exception):
    """An exception raised when attempting to subscribe to an Ethereum event
    without a subscription token.
    """
    pass


class GethRequest(BaseModel):
    """A class representing a JSON-RPC request to the Geth client."""
    jsonrpc: str = "2.0"
    """The version of the JSON-RPC protocol."""
    id: int | None = None
    """The ID of the request, or None if it is a notification."""
    method: str
    """The name of the RPC method to call."""
    params: list[Any]
    """A list of parameters to pass to the RPC method."""

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        frozen = True
        json_loads = orjson.loads
        json_dumps = json.orjson_dumps


class GethSuccessResponse(BaseModel):
    """A class representing a successful JSON-RPC response from the Geth
    client.
    """
    jsonrpc: str = "2.0"
    """The version of the JSON-RPC protocol."""
    id: int | None = None
    """The ID of the response, or None if it is a notification."""
    result: Any
    """The result of the RPC method call."""

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        frozen = True
        json_loads = orjson.loads
        json_dumps = json.orjson_dumps


class GethErrorDetail(BaseModel):
    """A class representing the error details in a JSON-RPC error response from
    the Geth client.
    """
    jsonrpc: str = "2.0"
    """The version of the JSON-RPC protocol."""
    code: int
    """The error code."""
    message: str
    """A description of the error."""

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        frozen = True
        json_loads = orjson.loads
        json_dumps = json.orjson_dumps


class GethError(Exception):
    """An exception representing an error response from the Geth client."""
    def __init__(self, error: GethErrorDetail | list[GethErrorDetail]) -> None:
        if isinstance(error, list):
            super().__init__([f"{err.code}: {err.message}" for err in error])
        else:
            self.code = error.code
            """The error code."""
            self.msg = error.message
            """A description of the error."""
            super().__init__(f"{error.code}: {error.message}")


class GethErrorResponse(BaseModel):
    """A class representing an error response from the Geth client."""
    jsonrpc: str = "2.0"
    """The version of the JSON-RPC protocol."""
    id: int | None
    """The ID of the response, or None if it is a notification."""
    error: GethErrorDetail
    """The details of the error."""

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        frozen = True
        json_loads = orjson.loads
        json_dumps = json.orjson_dumps


GethResponse = GethSuccessResponse | GethErrorResponse
"""A union class that represents the response data returned by a Geth node.

This class is a union of two other classes: ``GethSuccessResponse`` and
``GethErrorResponse``. Instances of ``GethSuccessResponse`` are returned when
a Geth request is successful, while instances of ``GethErrorResponse`` are
returned when a Geth request fails.
"""


class GethWSItem(BaseModel):
    """A class representing an item in a WebSocket response from the Geth
    client.
    """
    subscription: str
    """The subscription ID."""
    result: Any
    """The result of the subscription."""

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        frozen = True
        json_loads = orjson.loads
        json_dumps = json.orjson_dumps


class GethWSResponse(BaseModel):
    """A class representing a WebSocket response from the Geth client."""
    jsonrpc: str = "2.0"
    """The version of the JSON-RPC protocol."""
    method: str
    """The name of the WebSocket method."""
    params: GethWSItem
    """The item contained in the response."""

    @validator("params")
    def valid_params(cls, v: Any, **kwargs: Any) -> Any:
        if isinstance(v, str):
            if v != "eth_subscription":
                raise ValueError("Wrong WSResponse")
        return v

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        frozen = True
        json_loads = orjson.loads
        json_dumps = json.orjson_dumps


class CallOverrideParams(BaseModel):
    """A class representing override parameters for a contract function call.
    """
    balance: Wei | None = None
    """The balance of the caller."""
    nonce: Nonce | None = None
    """The nonce of the caller."""
    code: HexBytes | None = None
    """The code of the contract."""
    state: dict[str, Any] | None = None
    """The current state of the contract."""
    state_diff: dict[ChecksumAddress, dict[str, Any]] | None = \
        Field(None, alias="stateDiff")
    """The difference in the contract state."""

    def to_web3(self) -> Web3CallOverrideParams:
        """Converts the ``CallOverrideParams`` instance to an instance of
        ``web3.types.CallOverrideParams``.

        Returns:
            The converted ``web3.types.CallOverrideParams`` instance.
        """
        td = self.dict(by_alias=True, exclude_none=True)
        if "balance" in td:
            td["balance"] = typing.cast(Wei, td["balance"]).to_web3()
        if "code" in td:
            td["code"] = typing.cast(HexBytes, td["code"]).value
        return typing.cast(Web3CallOverrideParams, td)

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        frozen = True
        json_loads = orjson.loads
        json_dumps = json.orjson_dumps


CallOverride = dict[ChecksumAddress, CallOverrideParams]
"""A dictionary that maps Ethereum addresses to ``CallOverrideParams``
instances.

This class is used to specify the call override parameters for a smart
contract function. The keys of the dictionary are Ethereum addresses,
and the values are instances of the ``CallOverrideParams`` class, which
specify the call override parameters for the corresponding address.
"""


from .datatypes.base import (
    Address,
    BlockIdentifier,
    Gas,
    Hash32,
    HexBytes,
    IntStr,
    Wei,
)
from .datatypes.eth import (
    AccessEntry,
    AccessList,
    Block,
    FeeHistory,
    FilterParams,
    Log,
    Receipt,
    SyncStatus,
    Transaction,
    TxParams,
)
from .datatypes.geth import (
    CallOverride,
    CallOverrideParams,
    GethError,
    GethErrorDetail,
    GethErrorResponse,
    GethIsDead,
    GethRequest,
    GethResponse,
    GethSuccessResponse,
    GethWSItem,
    GethWSResponse,
    IdNotMatch,
    NoSubscribeToken,
)
from .datatypes.txpool import (
    TxpoolContent,
    TxpoolContentFrom,
    TxpoolInspect,
    TxpoolSnapshot,
    TxpoolStatus,
)

__all__ = [
    "Address",
    "BlockIdentifier",
    "Gas",
    "Hash32",
    "HexBytes",
    "IntStr",
    "Wei",
    "AccessEntry",
    "AccessList",
    "SyncStatus",
    "FeeHistory",
    "TxParams",
    "Transaction",
    "Block",
    "Log",
    "Receipt",
    "FilterParams",
    "CallOverride",
    "CallOverrideParams",
    "GethError",
    "GethErrorDetail",
    "GethErrorResponse",
    "GethIsDead",
    "GethRequest",
    "GethResponse",
    "GethSuccessResponse",
    "GethWSItem",
    "GethWSResponse",
    "IdNotMatch",
    "NoSubscribeToken",
    "TxpoolContent",
    "TxpoolContentFrom",
    "TxpoolInspect",
    "TxpoolSnapshot",
    "TxpoolStatus"
]
